#include "YCalcul.h"

///////////////////////////////////////////////////
// Fonctions de calcul
///////////////////////////////////////////////////


/**
 * Calcul la valeur entire la plus proche du rel pass en paramtre.
 *                |
 *               2+    ---
 *                |
 *               1+ ---
 *                |
 *     --+--+--+-=+=-+--+--+--
 *      -3 -2 -1  |  1  2  3
 *            --- +-1
 *                |
 * \param  x : la valeur rel  arrondir
 * \return La valeur arrondie
 */
short YCalcArrondi (double x) {
  if (x > 0)
    return (short)(x+0.5);
  else
    return (short)(x-0.5);
}

/**
 * Calcul le barycentre de l'objet obj pass en paramtre.
 * Si S1,...,Sn sont les sommets de obj, alors:
 *            1   n
 *   x(bar) = -.sigma {x(Si)}
 *            n  i=1
 *            1   n
 *   y(bar) = -.sigma {y(Si)}
 *            n  i=1
 *            1   n
 *   z(bar) = -.sigma {z(Si)}
 *            n  i=1
 * \param  obj : l'objet dont on veut calculer le barycentre
 * \return Les coordonnes du barycentre.
 */
point_3d YCalcBarycentre (objet_3d obj){
  point_3d bar;
  double somme_x, somme_y, somme_z;
  int i;

  for (i=0 ; i<obj.nb_sommets ; i++) {
    somme_x += (obj.sommets[i]).x;
    somme_y += (obj.sommets[i]).y;
    somme_z += (obj.sommets[i]).z;
  }
  bar.x = somme_x / obj.nb_sommets;
  bar.y = somme_y / obj.nb_sommets;
  bar.z = somme_z / obj.nb_sommets;

  return bar;
}

/**
 * Calcule l'quation du plan passant par les points A, B et C.
 * Les paramtres a, b, c et d sont remplis par les coefficients de cette quation.
 * \param A : Un point du plan
 * \param B : Un point du plan
 * \param C : Un point du plan
 * \param a : Le coefficient des 'x'
 * \param b : Le coefficient des 'y'
 * \param c : Le coefficient des 'z'
 * \param d : La constante
 */
void YCalcEqPlan (point_3d A, point_3d B, point_3d C, double *a, double *b, double *c, double *d) {
  double x1,y1,z1,x2,y2,z2;

  // Calcul du vecteur AB
  x1 = B.x - A.x;
  y1 = B.y - A.y;
  z1 = B.z - A.z;
  // Calcul du vecteur AC
  x2 = C.x - A.x;
  y2 = C.y - A.y;
  z2 = C.z - A.z;
  // Calcul de a, b et c (produit vectoriel de AB et AC)
  *a = y1*z2 - y2*z1;
  *b = z1*x2 - z2*x1;
  *c = x1*y2 - x2*y1;
  // Calcul de d
  *d = -(*a) * (A.x);
  *d -= (*b) * (A.y);
  *d -= (*c) * (A.z);
}

/**
 * Calcule l'quation de la face passe en paramtre. Les coefficients (a, b, c et d) de
 * l'quation sont stocks directement dans l'objet pass en paramtre.
 * L'quation de la face respecte la convention selon laquelle le vecteur normal (a,b,c)
 * de la face est orient vers l'intrieur de l'objet:
 *    Soit S(i) les sommets de la face indexs de sorte que pour i croissant le S(i) 'tournent'
 *    dans le sens qui 'oriente le vecteur normal n dans la bonne direction'.
 *           ->    ------->   ------->
 *    Alors, n  =  S(0)S(1) ^ S(0)S(j)
 *    avec j le plus petit indice tel que S(0), S(1) et S(j) ne soient pas colinaire.
 *    Puis d = -[a.x(S(0)) + b.y(S(0)) + c.z(S(0))]
 * \param  obj : l'objet dont la face fait parti
 * \param  numFace : l'indice (parmi toutes les faces de l'objet) de la face dont
 *                    on veut calculer l'quation 
 * \return  0 en cas de succs
 *         -1 sinon
 */
int YCalcEqFace (objet_3d obj, short numFace) {
  point_3d A,B;
  double a,b,c,d;
  int nbSommets;
  int i;

  // Vrifie que la face existe bien
  if (obj.nb_faces <= numFace) {
    fprintf(stderr,"Indice de face incorrecte dans YCalcEqFace\n");
    return -1;
  }
  // Init
  a=b=c=0;
  A.x = obj.sommets[((obj.faces[numFace]).sommets[0])].x;
  A.y = obj.sommets[((obj.faces[numFace]).sommets[0])].y;
  A.z = obj.sommets[((obj.faces[numFace]).sommets[0])].z;
  B.x = obj.sommets[((obj.faces[numFace]).sommets[1])].x;
  B.y = obj.sommets[((obj.faces[numFace]).sommets[1])].y;
  B.z = obj.sommets[((obj.faces[numFace]).sommets[1])].z;
  nbSommets = (obj.faces[numFace]).nb_sommets;
  // Calcul de a, b ,c et d
  for (i=2 ; IS_ABOUT_NUL(a)&&IS_ABOUT_NUL(b)&&IS_ABOUT_NUL(c) ; i++) {
    if (i == nbSommets) {
      // Tous les sommets sont situs sur la mme droite => erreur
      fprintf(stderr,"Impossible de calculer l'equation d'une face dont tous les pts sont sur une meme droite\n");
      return -1;
    }
    C.x = obj.sommets[((obj.faces[numFace]).sommets[i])].x;
    C.y = obj.sommets[((obj.faces[numFace]).sommets[i])].y;
    C.z = obj.sommets[((obj.faces[numFace]).sommets[i])].z;
    YCalcEqPlan(A,B,C,&a,&b,&c,&d);
  }
  // Sauvegarde dans la structure face_3d
  (obj.faces[numFace]).a = a;
  (obj.faces[numFace]).b = b;
  (obj.faces[numFace]).c = c;
  (obj.faces[numFace]).d = d;

  return 0;
}

/**
 * Calcul l'quation du plan, parallle  l'axe z dans le cas d'une vue en projection ou 
 * passant par l'origine dans le cas d'une perspective, passant par l'arte donne en paramtre .
 * \param  obj : l'objet dont l'arte fait parti
 * \param  numArete : l'indice (parmi toutes les artes de l'objet) de l'arte dont on
 *                    veut calculer l'quation du plan.
 * \param vue : Le type de vue souhait (en perspective ou projete).
 * \return  0 en cas de succs
 *         -1 sinon
 */
int YCalcEqPlanArete (objet_3d obj, short numArete, int vue) {
  double a,b,c,d;
  double xa,xb,ya,yb;
  point_3d A,B,C;

  // Vrifie que l'arte existe bien
  if (obj.nb_aretes <= numArete) {
    fprintf(stderr,"Indice d'arete incorrecte dans YCalcEqPlanArete\n");
    return -1;
  }
  // Calcul
  switch (vue) {
    case V_PROJ:
      xa = obj.sommets[((obj.aretes[numArete]).deb)].x;
      xb = obj.sommets[((obj.aretes[numArete]).fin)].x;
      ya = obj.sommets[((obj.aretes[numArete]).deb)].y;
      yb = obj.sommets[((obj.aretes[numArete]).fin)].y;
      a = yb - ya;
      b = xb - xa;
      c = 0;
      d = ya*(xb-xa)-xa*(yb-ya);
      break;
    case V_PERS:
      A.x = obj.sommets[((obj.aretes[numArete]).deb)].x;
      B.x = obj.sommets[((obj.aretes[numArete]).fin)].x;
      A.y = obj.sommets[((obj.aretes[numArete]).deb)].y;
      B.y = obj.sommets[((obj.aretes[numArete]).fin)].y;
      A.z = obj.sommets[((obj.aretes[numArete]).deb)].z;
      B.z = obj.sommets[((obj.aretes[numArete]).fin)].z;
      C.x = C.y = C.z = 0;
      YCalcEqPlan(A,B,C,&a,&b,&c,&d);
      break;
    default:
      fprintf(stderr,"Vue inconnue dans YCalcEqPlanArete.\n");
      return -1;
  }
  // Sauvegarde
  (obj.aretes[numArete]).a = a;
  (obj.aretes[numArete]).b = b;
  (obj.aretes[numArete]).c = c;
  (obj.aretes[numArete]).d = d;
  (obj.aretes[numArete]).eq_valide = (char)1;

  return 0;
}

/**
 * Vrifie si la face donne de l'objet pass en paramtre est du cot de l'observateur
 * ou non (si c'est une face avant ou non).
 * \param  obj : l'objet auquel la face appartient
 * \param  numFace : l'indice (parmi les autres faces de l'objet) de la face dont on veut 
 *                   tester la position.
 * \param vue : Le type de vue souhait (en perspective ou projete).
 * \return 0 si la face est arrire, 1 dans la cas contraire, -1 en cas d'erreur.
 */
int YCalcFaceAvant (objet_3d obj, short numFace, int vue) {
  int valO, valG;
  point_3d bar;

  // Vrifie que la face existe bien
  if (numFace >= obj.nb_faces) {
    fprintf(stderr,"Numero de face incconu dans YCalcFaceAvant.\n");
    return -1;
  }
  switch (vue) {
    case V_PROJ:
      // Comparaison (le cas 0 est la cas o on voit la face de profile)
      if ((obj.faces[numFace]).c < -ABOUT_NUL) // On veut : c < 0
        return 1;
      return 0;
    case V_PERS:
      // Calcul de G le barycentre
      bar = YCalcBarycentre(obj);
      // Calcul de F(O)
      valO = obj.faces[numFace].d;
      // Calcul de F(G)
      valG = obj.faces[numFace].a * bar.x;
      valG += obj.faces[numFace].b * bar.y;
      valG += obj.faces[numFace].c * bar.z;
      valG += obj.faces[numFace].d;
      // Comparaison (le cas 0 est la cas o on voit la face de profile)
      if ((valO*valG) < -ABOUT_NUL)
        return 1;
      return 0;
    default:
      fprintf(stderr,"Vue inconnue dans YCalcFaceAvant.\n");
      return -1;
  }
  return -1;
}

/** 
 * Vrifie pour toutes les faces de l'objet pass en paramtre si elles sont visibles ou 
 * non suivant qu'elles se trouvent ou non du cot de l'observateur.
 * Cette focntion fait les mmes oprations que YCalcFaceAvant mais en un seul appel de
 * fonction pour toutes les faces d'un objet (connomie de temps).
 * \param  obj : l'objet dont on veut connatre les faces visibles
 * \param vue : Le type de vue souhait (en perspective ou projete).
 * \return Un tableau comportant autant d'entres que l'objet possde de faces et 
 *         dont chaque case sera mise, pendant l'excution de la fonction, soit 
 *         0 si la face correspondante est cache, soit  1 dans la cas contraire.
 *         NULL en cas d'erreur.
 */
short* YCalcFacesAvant (objet_3d obj, int vue) {
  int i;
  int valO, valG;
  short nbFaces;
  short *faces;
  point_3d bar;

  nbFaces = obj.nb_faces;
  faces = calloc(nbFaces,sizeof(short));
  switch (vue) {
    case V_PROJ:
      for (i=0 ; i<nbFaces ; i++) {
	// Comparaison (le cas 0 est la cas o on voit la face de profile)
        if ((obj.faces[i]).c < -ABOUT_NUL)
          faces[i] = 1;
      }
      break;
    case V_PERS:
      // Calcul de G le barycentre
      bar = YCalcBarycentre(obj);
      for (i=0 ; i<nbFaces ; i++) {
        // Calcul de F(O)
        valO = obj.faces[i].d;
        // Calcul de F(G)
        valG = obj.faces[i].a * bar.x;
        valG += obj.faces[i].b * bar.y;
        valG += obj.faces[i].c * bar.z;
        valG += obj.faces[i].d;
	// Comparaison (le cas 0 est la cas o on voit la face de profile)
        if ((valO*valG) < -ABOUT_NUL)
          faces[i] = 1;
      }
      break;
    default:
      fprintf(stderr,"Vue inconnue dans YCalcFacesAvant.\n");
      free(faces);
      return NULL;
  }
  return faces;
}

int YCalcPointInt (point_3d pt, point_3d *polygone) {

}

/**
 * Calcule les segments visibles d'une droite par rapport  un objet grce  l'algorithme de Roberts.
 * La fonction se dcompose en 2 parties:
 *   - Pour chaque face de l'objet suceptible de cacher la droite, on effectue une srie de (au plus) 4
 *     tests pour voir si la face cache ou non la droite. Si une face cache un segment on l'ajoute 
 *      tous les segments cachs dj trouvs et on mmorise le tout dans une variable.
 *   - Une fois toutes les faces testes, on calcule,  partir du (l'objet tant convexe, il ne peut y
 *     en avoir qu'un seul) segment cach, la partie visible de la droite.
 * \param obj : L'objet suceptible de cacher la droite.
 * \param deb : Une extrmit de la droite dont on veut connatre les parties visibles.
 * \param fin : L'autre extrmit de la droite.
 * \param vue : Le type de vue souhait (en perspective ou projete) et s'il s'agit de la perspective
 *              le paramtre contient aussi la position du plan de perspective.
 * \return Un tableau de pointeurs sur des points. Chaque paire de points indique un
 *         segment visible de la droite. La dernire entre du tableau est mise  NULL
 *         pour signaler la fin du tableau.
 *         Ou NULL en cas de problme
 */
point_3d** YCalcAreteCachee (objet_3d obj, point_3d deb, point_3d fin, int vue) { // A faire...
  int i,j;
  int visible;
  short nbFaces,nbSommets,nbAretes;
  point_3d **segments_visibles;
  double *partie_cachee;
  double a,b,c,d;         // Coefficients de l'quation d'une face
  double q_a,q_b,q_d,q_c; // Coefficients de l'quation du plan Q passant par la droite et l'observateur
  double h_a,h_b,h_d,h_c; // Coefficients de l'quation du plan H passant par une arete et l'observateur
  int q_valide;           // Pour savoir si l'quation du plan Q est valide (1=vrai, 0=faux)
  double val,val_a,val_b,val_g;
  int somme;
  point_3d orig,bar;
  int pos_a,pos_b;
  double u,v,w,p,q,r,s,t,quotient;

  partie_cachee = NULL;
  q_valide = 0;
  // Teste le paramtre 'vue'
  switch (vue) {
    case V_PROJ:
    case V_PERS:
      break;
    default:
      fprintf(stderr,"Vue inconnue dans YCalcAreteCachee.\n");
      return NULL;
  }
  // Calcul de l'quation paramtrique de AB
  u = fin.x - deb.x;
  v = fin.y - deb.y;
  w = fin.z - deb.z;
  p = deb.x;
  q = deb.y;
  r = deb.z;
  // On regarde la visibilit de la droite (AB) par rapport  chaque face visible de l'objet
  nbFaces = obj.nb_faces;
  for (i=0 ; i<nbFaces ; i++) {
    if ((visible=YCalcFaceAvant(obj,i,vue)) == -1)
      return NULL;
    if (!visible) // La face n'est pas visible, on passe  la face suivante
      continue;
    /* Test 1 */
    // Coefficients de l'quation de la face F(x,y,z)
    a = (obj.faces[i]).a;
    b = (obj.faces[i]).b;
    c = (obj.faces[i]).c;
    d = (obj.faces[i]).d;
    if (vue == V_PROJ) { // cas d'une vue en projection
      if ((a*deb.x+b*deb.y+c*deb.z+d <= ABOUT_NUL) && ((a*fin.x+b*fin.y+c*fin.z+d <= ABOUT_NUL)))
	//(a*deb.x+b*deb.y+c*deb.z+d <= 0) && ((a*fin.x+b*fin.y+c*fin.z+d <= 0))
        continue; // Cette face ne masque pas la droite
    }
    else { // cas d'une vue en perspective
      // F(O) = d
      // Calcul de F(A)
      val_a = a*deb.x+b*deb.y+c*deb.z+d;
      // Calcul de F(B)
      val_a = a*fin.x+b*fin.y+c*fin.z+d;
      // Comparaison
      if (((val_a >= ABOUT_NUL) && (val_b >= ABOUT_NUL) && (d >= ABOUT_NUL)) || 
	  ((val_a <= ABOUT_NUL) && (val_b <= ABOUT_NUL) && (d <= ABOUT_NUL)))
	//((val_a >= 0) && (val_b >= 0) && (d >= 0)) || ((val_a <= 0) && (val_b <= 0) && (d <= 0))
        continue; // Cette face ne masque pas la droite
    }
    /* Test 2 */
    // Coefficients de l'quation du plan Q passant par la droite et l'observateur
    if (!q_valide) { // Pour ne pas recalculer plusieurs fois ces coeff (qui ne bouge pas suivant les faces)
      if (vue == V_PROJ) { // cas d'une vue en projection
        q_a = fin.y - deb.y;
        q_b = fin.x - deb.x;
        q_c = 0;
        q_d = deb.y*(fin.x-deb.x) - deb.x*(fin.y-deb.y);
      }
      else { // cas d'une vue en perspective
        orig.x = orig.y = orig.z = 0;
        YCalcEqPlan(deb,fin,orig,&q_a,&q_b,&q_c,&q_d);
      }
      q_valide = 1;
    }
    nbSommets = (obj.faces[i]).nb_sommets;
    if (IS_ABOUT_NUL(q_a) && IS_ABOUT_NUL(q_b) && IS_ABOUT_NUL(q_c)) {
      // Le test 2 ne peut traite le cas ou AB est vue de profile (correspond a 1 point)
      // Donc on passe tout de suite au test 3
    }
    else {
      // Position de chaque sommet par rapport  Q
      somme = 0;
      for (j=0 ; j<nbSommets ; j++) {
	// cas d'une vue en projection et en perspective
	// val = Q(S(i))
	val = q_a*(obj.sommets[((obj.faces[i]).sommets[j])].x);
	val += q_b*(obj.sommets[((obj.faces[i]).sommets[j])].y);
	val += q_c*(obj.sommets[((obj.faces[i]).sommets[j])].z);
	val += q_d;
	if (val > ABOUT_NUL)
	  somme ++;
	else if (val < -ABOUT_NUL)
	  somme --;
      }
      if ((somme < 2-nbSommets) || (nbSommets-2 < somme))
	continue; // Cette face ne masque pas la droite
    }
    /* Test 3 */
    somme = 0;
    // Calcul du barycentre
    bar = YCalcBarycentre(obj);
    // Position de la droite par rapport  chaque plan H[j] //  z passant par l'artes j de la face
    nbAretes = (obj.faces[i]).nb_aretes;
    for (j=0 ; j<nbAretes ; j++) {
      // Calcul de H[j]
      if (obj.aretes[((obj.faces[i]).aretes[j])].eq_valide == 0) {
        // Calcul 1 seule fois les coeff entre chaque transfo
        if (YCalcEqPlanArete(obj,(obj.faces[i]).aretes[j],vue) != 0)
          return NULL; // errreur!
      }
      h_a = (obj.aretes[((obj.faces[i]).aretes[j])].a);
      h_b = (obj.aretes[((obj.faces[i]).aretes[j])].b);
      h_c = (obj.aretes[((obj.faces[i]).aretes[j])].c);
      h_d = (obj.aretes[((obj.faces[i]).aretes[j])].d);
      if (IS_ABOUT_NUL(h_a) && IS_ABOUT_NUL(h_b) && IS_ABOUT_NUL(h_c)) {
	// Cas o l'arte j est //  z ou passe par O
	// On ignore cette face
      }
      else {
	// Valable pour une vue en projection et en perspective
	// Calcul de val=H[j](bar)
	val = h_a * bar.x;
	val += h_b * bar.y;
	val += h_c * bar.z;
	val += h_d;
	// Calcul de val_a=H[j](A)
	val_a = h_a * deb.x;
	val_a += h_b * deb.y;
	val_a += h_c * deb.z;
	val_a += h_d;
	// Calcul de val_b=H[j](B)
	val_b = h_a * fin.x;
	val_b += h_b * fin.y;
	val_b += h_c * fin.z;
	val_b += h_d;
	// Position de A par rapport  H[j]
	if (val_a*val < -ABOUT_NUL)
	  pos_a = 1; // A est extrieur du cylindre form par les plans H[j]
	else if (val_a*val > ABOUT_NUL)
	  pos_a = -1; // A est intrieur du cylindre form par les plans H[j]
	else
	  pos_a = 0; // A est sur le cylindre form par les plans H[j]
	// Position de B par rapport  H[j]
	if (val_b*val < -ABOUT_NUL)
	  pos_b = 1; // B est extrieur du cylindre form par les plans H[j]
	else if (val_b*val > ABOUT_NUL)
	  pos_b = -1; // B est intrieur du cylindre form par les plans H[j]
	else
	  pos_b = 0; // B est sur le cylindre form par les plans H[j]
	/* Sous-tests du test 3
	 * pos_a|pos_b| somme
	 *  -1    -1     -2
	 *  -1     0     -1
	 *  -1     1      0
	 *   0    -1     -1
	 *   0     0      0
	 *   0     1      1
	 *   1    -1      0
	 *   1     0      1
	 *   1     1      2
	 */
	if ((pos_a + pos_b) > 0) {
	  // Un des points est ext et l'autre est soit ext soit sur H[j]
	  // Alors AB n'est pas masqu par la face i
	  somme = -1;
	  break;
	}
	if ((pos_a + pos_b) == 0) {
	  // Un des points est ext et l'autre int ou les 2 points sont sur H[j]
	  // On ne peut pas dcider -> test 4
	  break;
	}
      }
      somme ++;
    }
    if (somme == -1)
      continue; // Cette face ne masque pas la droite
    else if (somme == nbAretes) {
      // L'arte est entirement masque par la face
      if (partie_cachee == NULL) {
	// La face courante est la 1ere face qui cache une partie de la droite
	partie_cachee = malloc(2*sizeof(double));
      }
      partie_cachee[0] = 0; // t=0
      partie_cachee[1] = 1; // t=1
      break;
    }
    /* Test 4 */
    somme = 0;
    // Intersection de AB avec chaque plan H[j]
    for (j=0 ; j<nbAretes ; j++) {
      // Calcul de H[j]
      if (obj.aretes[((obj.faces[i]).aretes[j])].eq_valide == 0) {
        // Calcul 1 seule fois les coeff entre chaque transfo
        if (YCalcEqPlanArete(obj,(obj.faces[i]).aretes[j].vue) != 0)
          return NULL; // errreur!
      }
      h_a = (obj.aretes[((obj.faces[i]).aretes[j])].a);
      h_b = (obj.aretes[((obj.faces[i]).aretes[j])].b);
      h_c = (obj.aretes[((obj.faces[i]).aretes[j])].c);
      h_d = (obj.aretes[((obj.faces[i]).aretes[j])].d);
      // Calcul de l'intersection entre AB et H[j]
      quotient = u * h_a;
      quotient += v * h_b;
      quotient += w * h_c;
      if (!IS_ABOUT_NUL(quotient)) {
	val = p * h_a;
	val += q * h_b;
	val += r * h_c;
	val += h_d;
	t = -val / quotient;
	// Pour les imprcision de calcul
	if (IS_ABOUT_NUL(t))
	  t = 0;
	else if (IS_ABOUT_NUL(t-1))
	  t = 1;
      }
      else
	t = -1;
      // Sous-tests du test 4
      if ((t>=0) && (t<=1)) {
	somme ++;
	// Calcul de F(I) avec I le point d'intersection
	val = a*(u*t+p) + b*(v*t+q) + c*(w*t+r) + d;
	// Calcul de F(bar)
	val_g = a*bar.x + b*bar.y + c*bar.z + d;
	if (val*val_g < -ABOUT_NUL) { // F(I)*F(bar)<0
	  // I est devant la face
	  // donc AB est devant la face (ie visible)
	  somme = -1; // Passe  la face suivante
	  break;
	}
	else if (val*val_g > ABOUT_NUL) { // F(I)*F(bar)>0
	  // I est derrire la face
	  if (somme == 1) { // C'est le 1er point d'intersection entre AB et un plan H[k]
	    s = t;
	    // Calcul de val=H[j](bar)
	    val = h_a * bar.x;
	    val += h_b * bar.y;
	    val += h_c * bar.z;
	    val += h_d;
	    // Calcul de val_a=H[j](A)
	    val_a = h_a * deb.x;
	    val_a += h_b * deb.y;
	    val_a += h_c * deb.z;
	    val_a += h_d;
	    // Position de A par rapport  H[j] et au barycentre
	    pos_a = val * val_a;
	  }
	  else { // C'est le deuxime point d'intersection entre AB et un plan H[k]
	    if (partie_cachee == NULL) { // Cette face est la 1ere face qui cache une partie de la droite
	      partie_cachee = malloc(2*sizeof(double));
	      if (s < t) {
		partie_cachee[0] = s;
		partie_cachee[1] = t;
	      }
	      else {
		partie_cachee[0] = t;
		partie_cachee[1] = s;
	      }
	    }
	    else { // La face courante n'est pas la 1ere face qui cache une partie de la droite
	      if (s < partie_cachee[0])
		partie_cachee[0] = s;
	      if (t < partie_cachee[0])
		partie_cachee[0] = t;
	      if (s > partie_cachee[1])
		partie_cachee[1] = s;
	      if (t > partie_cachee[1])
		partie_cachee[1] = t;
	    }
	    break; // Passe  la face suivante
	  }
	}
      }
    }
    if (somme != 1) // Cas dj traits...
      continue;
    else { // Il n'y a qu'un seul point d'intersection entre AB et un plan H[k] (Cas restant  trait)
      // Teste la visibilit de A
      if (pos_a < -ABOUT_NUL) { // A est visible 
	// On cache la partie de AB entre I et B
	if (partie_cachee == NULL) { // Cette face est la 1ere face qui cache une partie de la droite
	  partie_cachee = malloc(2*sizeof(double));
	  partie_cachee[0] = s;
	  partie_cachee[1] = 1;
	  continue;
	}
	else { // La face courante n'est pas la 1ere face qui cache une partie de la droite
	  if (s < partie_cachee[0])
	    partie_cachee[0] = s;
	  partie_cachee[1] = 1;
	}
      }
      else { // A est non visible
	// On cache la partie de AB entre A et I
	if (partie_cachee == NULL) { // Cette face est la 1ere face qui cache une partie de la droite
	  partie_cachee = malloc(2*sizeof(double));
	  partie_cachee[0] = 0;
	  partie_cachee[1] = s;
	  continue;
	}
	else { // La face courante n'est pas la 1ere face qui cache une partie de la droite
	  partie_cachee[0] = 0;
	  if (s > partie_cachee[1])
	    partie_cachee[1] = s;
	}
      }
    }
  } // Fin du traitement de toutes les faces de l'objet
  // Reconstitution de la ou des parties visibles de la droite
  segments_visibles = malloc(5*sizeof(point_3d*));
  if (partie_cachee == NULL) { // Droite entirement visible
    segments_visibles[0] = malloc(sizeof(point_3d));
    segments_visibles[1] = malloc(sizeof(point_3d));
    segments_visibles[2] = NULL;
    segments_visibles[0]->x = deb.x;
    segments_visibles[0]->y = deb.y;
    segments_visibles[0]->z = deb.z;
    segments_visibles[1]->x = fin.x;
    segments_visibles[1]->y = fin.y;
    segments_visibles[1]->z = fin.z;
  }
  else { // La droite est au moins en parti cache
    // Pour rectifier les imprcision de calcul
    if (IS_ABOUT_NUL(partie_cachee[0]))
      partie_cachee[0] = 0;
    if (IS_ABOUT_NUL(partie_cachee[1]-1))
      partie_cachee[1] = 1;
    // Calculs...
    if ((partie_cachee[0] == 0) && (partie_cachee[1] == 1)){// Cas particulier de la droite entirement cache
      segments_visibles[0] = NULL;
    }
    else if ((partie_cachee[0] == 0) || (partie_cachee[1] == 1)){// Cas o un seul segment est visible
      segments_visibles[0] = malloc(sizeof(point_3d));
      segments_visibles[1] = malloc(sizeof(point_3d));
      segments_visibles[2] = NULL;
      // Calcul des coordonnes des extrmits
      if (partie_cachee[0]  == 0) {// Segment visible entre I et B
	segments_visibles[0]->x = u*partie_cachee[1]+p;
	segments_visibles[0]->y = v*partie_cachee[1]+q;
	segments_visibles[0]->z = w*partie_cachee[1]+r;
	segments_visibles[1]->x = fin.x;
	segments_visibles[1]->y = fin.y;
	segments_visibles[1]->z = fin.z;
      }
      else { // Segment visible entre A et I
	segments_visibles[0]->x = deb.x;
	segments_visibles[0]->y = deb.y;
	segments_visibles[0]->z = deb.z;
	segments_visibles[1]->x = u*partie_cachee[0]+p;
	segments_visibles[1]->y = v*partie_cachee[0]+q;
	segments_visibles[1]->z = w*partie_cachee[0]+r;
      }
    }
    else { // Cas o deux segments sont visibles
      segments_visibles[0] = malloc(sizeof(point_3d));
      segments_visibles[1] = malloc(sizeof(point_3d));
      segments_visibles[2] = malloc(sizeof(point_3d));
      segments_visibles[3] = malloc(sizeof(point_3d));
      segments_visibles[4] = NULL;
      // Calcul des coord du 1er segment
      segments_visibles[0]->x = deb.x;
      segments_visibles[0]->y = deb.y;
      segments_visibles[0]->z = deb.z;
      segments_visibles[1]->x = u*partie_cachee[0]+p;
      segments_visibles[1]->y = v*partie_cachee[0]+q;
      segments_visibles[1]->z = w*partie_cachee[0]+r;
      // Calcul des coord du 2eme segment
      segments_visibles[2]->x = u*partie_cachee[1]+p;
      segments_visibles[2]->y = v*partie_cachee[1]+q;
      segments_visibles[2]->z = w*partie_cachee[1]+r;
      segments_visibles[3]->x = fin.x;
      segments_visibles[3]->y = fin.y;
      segments_visibles[3]->z = fin.z;
    }
  }

  return segments_visibles;
}

